/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/


package com.piercefuller.a7trk.decode;

import java.io.*;

    /// <summary>
    /// Summary description for AdcSource.
    /// </summary>
    public class CrowdSource implements Source
    {
        private final String Name = "Cr";

        private Source source;
        private long lastPosition = 0;
        private long limit = 0;
        private int nchannel = 0;
        private int nlevel = 0;
        private int[][][] history;
        private int historyIndex;
        private int crowd;
        private int setup;
        private int[] tcount;
        private int[] before;
        private int[] after;

        private BufferedWriter log = null; //new IO.streamWriter("c:\\tmp\\crowd.log");
        private int setsn = 0;

        public CrowdSource(Source s)
        {
            source = s;
            if (source == null) 
            {
                throw new Error("CrowdSource has no source");
            }
            limit = s.getEnd();

            nchannel = s.getNumChannels();
            tcount = new int[nchannel];
            before = new int[nchannel];
            after = new int[nchannel];
            for (int ch = 0; ch < nchannel; ch++) 
            {
                tcount[ch] = 0;
                before[ch] = 0;
                after[ch] = 0;
            }

            nlevel = 1;
            for (Source sl = source; sl != null; nlevel++, sl = sl.getSource());
            checkParameters(0);
        }

        private void checkParameters(long pos)
        {
            crowd = getTapeParameters(pos).getCrowd();
            if (crowd > 0)
            {
                int hsize = crowd+5;
                if (history == null || history.length < hsize) 
                {
                    history = new int[hsize][][];
                    for (int hi = 0; hi < history.length; hi++)
                    {
                        history[hi] = new int[nlevel][];
                        for (int level = 0; level < nlevel; level++) 
                        {
                            history[hi][level] = new int[nchannel];
                            for (int ch = 0; ch < nchannel; ch++) 
                            {
                                history[hi][level][ch] = 0;
                            }
                        }
                    }
                }
            }
            historyIndex = 0;
        }

        public TapeParameters getTapeParameters() { return getTapeParameters(getPosition()); }
        public TapeParameters getTapeParameters(long pos) { return source.getTapeParameters(pos); }
                
        public String getName() { return Name; }

        public Source getSource() { return source; }

        public int getNumChannels() { return source.getNumChannels(); }

        public String getChannelName(int channel) { return source.getChannelName(channel); }

        public long getStart() { return source.getStart(); }

        public long getEnd() { return source.getEnd(); }

        public void setStart(long pos) { source.setStart(pos); }

        public void setEnd(long pos) { source.setEnd(pos); }

        public long getLength() { return source.getLength(); }

        public long getPosition()
        {
            if (crowd > 0)
            {
                return source.getPosition() - crowd - 1;
            }
            else
            {
                return source.getPosition();
            }
        }

        public void setPosition(long pos)
        {
            if (log != null) 
            {
                try {
					log.write(pos + ". Set " + ++setsn);
					log.newLine();
				} catch (IOException e) {
					System.err.println(e);
				}
            }
            checkParameters(pos);
            lastPosition = pos;
            if (crowd > 0)
            {
                source.setPosition(pos - crowd);
                for (int ch = 0; ch < nchannel; ch++) 
                {
                    tcount[ch] = 0;
                    before[ch] = crowd+1;
                    after[ch] = crowd+1;
                    for (int level = 0; level < history[0].length; level++)
                    {
                        for (int hi = 0; hi < history.length; hi++) 
                        {
                            history[hi][level][ch] = 0;
                        }
                    }
                }
                setup = crowd;
                for (; getPosition() < pos; ) 
                {
                    advance();
                }
            }
            else
            {
                source.setPosition(pos);
            }
        }

        public void resetPosition()
        {
            setPosition(lastPosition);
        }

        public long getLimit()
        {
            return limit;
        }

        public void setLimit(long l)
        {
            limit = l;
        }

        public void resetLimit()
        {
            limit = source.getEnd();
        }

        public boolean atLimit()
        {
            return getPosition() >= limit;
        }

        public void advance()
        {
            if (crowd > 0)
            {
                for (int level = 1; level < nlevel; level++)
                {
                    int[] samples = source.getSamples(level-1);
                    for (int ch = 0; ch < nchannel; ch++) 
                    {
                        history[historyIndex][level][ch] = samples[ch];
                    }
                }

                int pick = 0;
                for (int i = 0; i < nchannel; i++) 
                {
                    int hc = historyIndex - crowd;
                    if (hc < 0) 
                    {
                        hc += history.length;
                    }
                    pick = 0;
                    if (before[i] < crowd) 
                    {
                        pick++;
                        if (before[i] < crowd/2) 
                        {
                            pick++;
                        }
                    }
                    if (after[i] < crowd)
                    {
                        pick--;
                        if (after[i] < crowd/2)
                        {
                            pick--;
                        }
                    }
                    int hp = hc + pick;
                    if (hp < 0) 
                    {
                        hp += history.length;
                    } 
                    else if (hp >= history.length) 
                    {
                        hp -= history.length;
                    }
                    history[hc][0][i] = history[hp][1][i];

                    hc = historyIndex - crowd - 3;
                    if (hc < 0) 
                    {
                        hc += history.length;
                    }
                    int hr = hc - 1;
                    if (hr < 0) 
                    {
                        hr += history.length;
                    }
                    if (setup > 0) 
                    {
                        if (i == nchannel-1) 
                        {
                            setup--;
                        }
                    } 
                    else 
                    {
                        if (history[hc][1][i] != 0 && history[hc][1][i] != history[hr][1][i]) 
                        {
                            tcount[i]--;
                            before[i] = after[i];
                            if (tcount[i] > 1) 
                            {
                                hc += after[i];
                                after[i] = 1;
                                for (;;) 
                                {
                                    hc++;
                                    if (hc >= history.length) 
                                    {
                                        hc -= history.length;
                                    }
                                    if (hc == historyIndex) 
                                    {
                                        tcount[i] = 1;
                                        break;
                                    }
                                    hr = hc - 1;
                                    if (hr < 0) 
                                    {
                                        hr += history.length;
                                    }
                                    if (history[hc][1][i] != 0 && history[hc][1][i] != history[hr][1][i]) 
                                    {
                                        break;
                                    }
                                    after[i]++;
                                }
                            }
                            else if (tcount[i] == 1)
                            {
                                after[i] = crowd + 3 - after[i];
                            }
                            else
                            {
                                after[i] = crowd + 3;
                            }
                        }
                    }

                    hc = historyIndex;
                    hr = historyIndex-1;
                    if (hr < 0) 
                    {
                        hr += history.length;
                    }
                    if (history[hc][1][i] != 0 && history[hc][1][i] != history[hr][1][i]) 
                    {
                        tcount[i]++;
                        if (tcount[i] == 1) 
                        {
                            after[i] = 1;
                        }
                    } 
                    else if (tcount[i] == 1) 
                    {
                        after[i]++;
                    }
                }

                if (log != null)
                {
                    try {
						log.write(getPosition() + "." + setsn + (setup > 0? "* " : "  ") +
						    before[6] + " " + after[6] + " " + tcount[6] + " " + pick + ": " +
						    getSamples(0)[6] + " " +
						    getSamples(1)[6] + " " +
						    getSamples(2)[6] + " " +
						    getSamples(3)[6] + " " +
						    getSamples(4)[6] + "   ");
						log.newLine();
						log.flush();
					} catch (IOException e) {
						System.err.println(e);
					}
                }

                historyIndex++;
                if (historyIndex == history.length) 
                {
                    historyIndex = 0;
                }
            }
            source.advance();
        }

        public int[] getSamples()
        {
            return getSamples(0);
        }

        public int[] getSamples(int level)
        {
            if (crowd > 0)
            {
                int hc = historyIndex - crowd - 1;
                if (hc < 0) 
                {
                    hc += history.length;
                }
                return history[hc][level];
            }
            if (level == 0) 
            {
                return source.getSamples();
            }
            return source.getSamples(level-1);
        }

        public int getFlags()
        {
            return source.getFlags();
        }

        public void close()
        {
            if (source != null) 
            {
                source.close();
            }
            source = null;
        }
    }
