/**
 * DatRawSource.java
 *
 * Copyright (C) 2013 Paul R. Pierce
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see http://www.gnu.org/licenses/gpl.txt
 */
package com.piercefuller.a7trk.decode;

import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * 
 */
public class DatRawSource extends RawSource {
    private static final int    Channels     = 7;
    private static final String ChannelNames = "0123456P";

    private RandomAccessFile    file         = null;
    private final long          length;
    private final int           sampleSize   = 16;
    private byte[]              buf          = new byte[sampleSize];
    private boolean             phase        = false;

    /**
     * @param datafile
     * @throws Exception
     */
    public DatRawSource(final DataFile datafile) throws Exception {
        super(datafile);
        file = new RandomAccessFile(datafile.getFilename(), "r");
        length = 2 * file.length() / sampleSize;
        initialize();
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#getLength()
     */
    public long getLength() {
        return length;
    }

    @Override
    public Object clone() {
        try {
            return new DatRawSource(getDatafile());
        } catch (Exception e) {
            System.err.println(e);
            return null;
        }
    }

    @Override
    public void readFlat(final int[] samples) throws Exception {
        if (file == null) {
            throw new Error("Source closed");
        }
        phase = (getPosition() & 1) != 0;
        int p = 0;
        for (int ch = 0; ch < Channels; ch++) {
            int v = (buf[p++] & 0xFF);
            v += (buf[p++]) << 8;
            v -= 0x0800;
            samples[ch] = v;
        }
        if (phase) {
            file.read(buf, 0, sampleSize);
            p = 0;
            for (int ch = 0; ch < Channels; ch++) {
                int v = (buf[p++] & 0xFF);
                v += (buf[p++]) << 8;
                v -= 0x0800;
                samples[ch] = (samples[ch] + v) / 2;
            }
            phase = false;
        } else {
            phase = true;
        }
    }

    public int getNumChannels() {
        return Channels;
    }

    public String getChannelName(final int channel) {
        return ChannelNames.substring(channel, channel + 1);
    }

    public int getFlags() {
        int v = (buf[sampleSize - 2] & 0xFF);
        v += (buf[sampleSize - 1]) << 8;
        return ~v;
    }

    public void close() {
        if (file != null) {
            try {
                file.close();
            } catch (Exception e) {
                System.err.println(e);
            }
        }
        file = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.RawSource#sampleSeek(int)
     */
    @Override
    public void sampleSeek(final long n) throws IOException {
        if (file == null) {
            throw new Error("Source closed");
        }
        file.seek(sampleSize * n);
        file.read(buf, 0, sampleSize);
    }

}
