/**********************************
 *
 * Copyright (C) 2004,2013 Paul Pierce
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

import java.io.BufferedWriter;
import java.io.IOException;

public class DataFile {
    private String             artifactid;
    private String             filename;
    private int                nchannel;
    private long               start;
    private long               end;
    private RecordList         records;
    private TapeParametersList tpl;

    public String getArtifactId() {
        return artifactid;
    }

    public void setArtifactId(final String value) {
        artifactid = value;
    }

    public String getFilename() {
        return filename;
    }

    public void setFilename(final String value) {
        filename = value;
    }

    public int getChannels() {
        return nchannel;
    }

    public void setChannels(final int value) {
        nchannel = value;
    }

    public long getStart() {
        return start;
    }

    public void setStart(final long value) throws Exception {
        start = value;
        init_tpl();
    }

    public long getEnd() {
        return end;
    }

    public void setEnd(final long value) throws Exception {
        end = value;
        init_tpl();
    }

    public RecordList getRecords() {
        return records;
    }

    public TapeParametersList getTapeParameters() {
        return tpl;
    }

    public DataFile(final String filename, final int nchannel, final int start, final int end) throws Exception {
        artifactid = null;
        this.filename = filename;
        this.nchannel = nchannel;
        this.start = start;
        this.end = end;
        records = new RecordList();
        tpl = new TapeParametersList(nchannel);
        init_tpl();
    }

    public void write(final BufferedWriter ow, final int indent) throws IOException {
        int direction = 0;
        for (int i = 0; i < records.getCount(); i++) {
            if (records.get(i).getMotionForward()) {
                direction |= 1;
            } else {
                direction |= 2;
            }
        }
        String sdir = "";
        if (direction == 1) {
            sdir = " direction=\"forward\"";
        }
        if (direction == 2) {
            sdir = " direction=\"reverse\"";
        }

        String inds = "                    ".substring(0, indent);
        if (artifactid != null) {
            ow.write("<datafile artifact=\"" + artifactid + "\"");
            ow.newLine();
        } else {
            ow.write("<datafile");
            ow.newLine();
        }
        ow.write(inds + " filename=\"" + filename + "\"" + " form=\"tape\"" + " channels=\"" + nchannel + "\"" + sdir + " start=\"" + start + "\"" + " end=\""
                + end + "\">");
        ow.newLine();

        int tc = tpl.getCount();
        int rc = records.getCount();
        long lostart = start;
        long loend = start;
        long histart, hiend;
        for (; tc > 0 || rc > 0;) {
            Object best = null;
            histart = end;
            hiend = end;
            for (int i = 0; i < tpl.getCount(); i++) {
                TapeParameters tp = tpl.get(i);
                if (tp.getStart() >= lostart && tp.getStart() <= histart && (tp.getStart() > lostart || tp.getEnd() > loend)
                        && (tp.getStart() < histart || tp.getEnd() < hiend)) {
                    histart = tp.getStart();
                    hiend = tp.getEnd();
                    best = tp;
                }
            }
            for (int i = 0; i < records.getCount(); i++) {
                Record rec = records.get(i);
                if (rec.getStart() >= lostart && rec.getStart() <= histart && (rec.getStart() > lostart || rec.getEnd() > loend)
                        && (rec.getStart() < histart || rec.getEnd() < hiend)) {
                    histart = rec.getStart();
                    hiend = rec.getEnd();
                    best = rec;
                }
            }
            lostart = histart;
            loend = hiend;
            if (best == null) {
                break;
            }
            if (best instanceof TapeParameters) {
                ((TapeParameters)best).write(ow, indent + 2);
                tc--;
                for (int i = 0; i < records.getCount(); i++) {
                    Record rec = records.get(i);
                    if (rec.getStart() == lostart && rec.getEnd() == loend) {
                        best = rec;
                        break;
                    }
                }
            }
            if (best instanceof Record) {
                ((Record)best).write(ow, indent + 2);
                rc--;
            }
        }
        ow.write(inds + "</datafile>");
        ow.newLine();
    }

    private void init_tpl() throws Exception {
        if (tpl.getCount() == 0 && start >= 0 && end >= 0) {
            tpl.add(start, end);
        }
    }
}
