/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/


package com.piercefuller.a7trk.decode;

    /// <summary>
    /// Summary description for AdcSource.
    /// </summary>
    public class DirSource implements Source
    {
        private final String Name = "Dir";

        private Source source;
        private long lastPosition = 0;
        private long limit = 0;
        private int nchannel = 0;
        private int nlevel = 0;
        private int[][] history;
        private int[] max;
        private int[] min;

        public DirSource(Source s)
        {
            source = s;
            if (source == null) 
            {
                throw new Error("DirSource has no source");
            }
            if (!(source instanceof BaselineSource)) 
            {
                throw new Error("DirSource must be based on BaselineSource");
            }
            limit = s.getEnd();

            nchannel = s.getNumChannels();
            max = new int[nchannel];
            min = new int[nchannel];
            for (int ch = 0; ch < nchannel; ch++) 
            {
                max[ch] = 0;
                min[ch] = 0;
            }

            nlevel = 1;
            for (Source sl = source; sl != null; nlevel++, sl = sl.getSource());

            history = new int[nlevel][];
            for (int level = 0; level < nlevel; level++) 
            {
                history[level] = new int[nchannel];
                for (int ch = 0; ch < nchannel; ch++) 
                {
                    history[level][ch] = 0;
                }
            }
        }

        public TapeParameters getTapeParameters() { return getTapeParameters(getPosition()); }
        public TapeParameters getTapeParameters(long pos) { return source.getTapeParameters(pos); }
        
        public String getName() { return Name; }

        public Source getSource() { return source; }

        public int getNumChannels() { return source.getNumChannels(); }

        public String getChannelName(int channel) { return source.getChannelName(channel); }

        public long getStart() { return source.getStart(); }

        public long getEnd() { return source.getEnd(); }

        public void setStart(long pos) { source.setStart(pos); }

        public void setEnd(long pos) { source.setEnd(pos); }

        public long getLength() { return source.getLength(); }

        public long getPosition()
        {
            return source.getPosition() - 1;
        }

        public void setPosition(long pos)
        {
            lastPosition = pos;
            source.setPosition(pos);
            for (int ch = 0; ch < nchannel; ch++) 
            {
                max[ch] = 0;
                min[ch] = 0;
                for (int level = 0; level < history.length; level++)
                {
                    history[level][ch] = 0;
                }
            }
            advance();
        }

        public void resetPosition()
        {
            setPosition(lastPosition);
        }

        public long getLimit()
        {
            return limit;
        }

        public void setLimit(long l)
        {
            limit = l;
        }

        public void resetLimit()
        {
            limit = source.getEnd();
        }

        public boolean atLimit()
        {
            return source.getPosition() - 1 >= limit;
        }

        public void advance()
        {
            boolean lowClip = ((BaselineSource)source).inBaseline(getPosition());
            for (int level = 1; level < nlevel; level++)
            {
                int[] samples = source.getSamples(level-1);
                for (int ch = 0; ch < nchannel; ch++) 
                {
                    history[level][ch] = samples[ch];
                }
            }

            for (int i = 0; i < nchannel; i++) 
            {
                int dir = history[0][i];
                int samp = history[1][i];
                if (samp < 0) 
                {
                    max[i] = 0;
                    if (dir > 0) 
                    {
                        dir = 0;
                    }
                } 
                else if (samp >= max[i]) 
                {
                    max[i] = samp;
                }
                else if (max[i] > getTapeParameters().getClip(lowClip, i)) 
                {
                    dir = max[i];
                } 
                if (samp > 0) 
                {
                    min[i] = 0;
                    if (dir < 0) 
                    {
                        dir = 0;
                    }
                } 
                else if (samp < min[i]) 
                {
                    min[i] = samp;
                } 
                else if (min[i] < -getTapeParameters().getClip(lowClip, i)) 
                {
                    dir = min[i];
                }
                history[0][i] = dir;
            }
            source.advance();
        }

        public int[] getSamples()
        {
            return getSamples(0);
        }

        public int[] getSamples(int level)
        {
            return history[level];
        }

        public int getFlags()
        {
            return source.getFlags();
        }

        public void close()
        {
            if (source != null) 
            {
                source.close();
            }
            source = null;
        }
    }
