/**
 * HpfRawSource.java
 *
 * Copyright (C) 2013 Paul R. Pierce
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see http://www.gnu.org/licenses/gpl.txt
 */
package com.piercefuller.a7trk.decode;

import java.io.IOException;

/**
 * 
 */
public class HpfRawSource extends RawSource {

    private static final String ChannelNames   = "0123456789";

    private HpfFile             hpfFile;
    private final long          length;
    private final int           channels;
    private final int[]         altSamples;
    private long                samplePosition = 0;

    /**
     * @param datafile
     * @throws Exception
     */
    public HpfRawSource(final DataFile datafile) throws Exception {
        super(datafile);
        hpfFile = new HpfFile(datafile.getFilename());
        length = 2 * hpfFile.getTotalSamples();
        channels = hpfFile.getChannels();
        altSamples = new int[channels];
        initialize();
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#getNumChannels()
     */
    public int getNumChannels() {
        return channels;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#getChannelName(int)
     */
    public String getChannelName(final int channel) {
        return ChannelNames.substring(channel, channel + 1);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#getLength()
     */
    public long getLength() {
        return length;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#getFlags()
     */
    public int getFlags() {
        return 0;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.Source#close()
     */
    public void close() {
        if (hpfFile != null) {
            try {
                hpfFile.close();
            } catch (Exception e) {
                System.err.println(e);
            }
        }
        hpfFile = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.RawSource#clone()
     */
    @Override
    public Object clone() {
        try {
            return new HpfRawSource(getDatafile());
        } catch (Exception e) {
            System.err.println(e);
            return null;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.RawSource#sampleSeek(long)
     */
    @Override
    public void sampleSeek(final long n) throws IOException {
        if (hpfFile == null) {
            throw new Error("Source closed");
        }
        samplePosition = n;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.piercefuller.a7trk.decode.RawSource#readFlat(int[])
     */
    @Override
    public void readFlat(final int[] samples) throws Exception {
        if (hpfFile == null) {
            throw new Error("Source closed");
        }
        hpfFile.getSamples(samples, samplePosition, 0, getTapeParameters().getGain());
        if ((getPosition() & 1) != 0) {
            ++samplePosition;
            hpfFile.getSamples(altSamples, samplePosition, 0, getTapeParameters().getGain());
            for (int ch = 0; ch < channels; ch++) {
                samples[ch] = (samples[ch] + altSamples[ch]) / 2;
            }
        }
    }
}
