/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/

package com.piercefuller.a7trk.decode;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

/// <summary>
/// Summary description for TapeDecoder.
/// </summary>
public class TapeDecoder implements MultiSource {
    private final int       ClumpSize     = 10000;
    private final int       MaxRecord     = 100 * 32768 * 6;
    private final int       NominalRecord = 20 * 50000;
    private final int       Pad           = 50;

    private DataCollection  datacollection;
    private DataFile        datafile;
    private Source          source;
    private int             nchannel;
    private int             evalErrors;
    private int             evalFixups;
    private int             gapJunk;
    private int[]           spacingHist;
    private ProgressWatcher progressWatcher;
    private RecordInfo      ri;

    private Record          currentRecord;
    private RecordInfo      currentRecordInfo;

    public TapeDecoder(final String filename) throws Exception {
        int dot = filename.lastIndexOf('.');
        if (dot >= 0 && filename.indexOf(File.pathSeparator, dot) > dot) {
            dot = -1;
        }
        if (dot < 0) {
            try {
                openXml(filename + ".xml");
            } catch (IOException e1) {
                System.err.println(e1);
                try {
                    openData(filename + ".dat");
                } catch (IOException e2) {
                    System.err.println(e2);
                    openData(filename);
                }
            }
        } else if (filename.toLowerCase().endsWith(".xml")) {
            openXml(filename);
        } else {
            openData(filename);
        }
    }

    private void openXml(final String filename) throws Exception {
        datacollection = new DataCollection(filename);
        datafile = datacollection.getDataFile();
        initialize(RawSource.getInstance(datafile));
        for (int i = 0; i < datafile.getRecords().getCount(); i++) {
            evalErrors += datafile.getRecords().get(i).getErrors();
            evalFixups += datafile.getRecords().get(i).getFixups();
        }
    }

    private void openData(final String filename) throws Exception {
        datafile = new DataFile(filename, 7, -1, -1);
        initialize(RawSource.getInstance(datafile));
        datacollection = new DataCollection(datafile);
    }

    private TapeDecoder(Source s, final DataCollection dc, final DataFile datafile) {
        this.datafile = datafile;
        datacollection = dc;
        for (;;) {
            if (s instanceof RawSource) {
                break;
            }
            s = s.getSource();
            if (s == null) {
                throw new Error("TapeDecoder must be based on RawSource");
            }
        }
        initialize((RawSource)((RawSource)s).clone());
    }

    private void clearStats(final boolean clearJunk) {
        evalErrors = 0;
        evalFixups = 0;
        if (clearJunk) {
            gapJunk = 0;
        }
        if (spacingHist != null) {
            for (int i = 0; i < spacingHist.length; i++) {
                spacingHist[i] = 0;
            }
        }
    }

    private void initialize(final RawSource rawSource) {
        nchannel = rawSource.getNumChannels();
        ri = new RecordInfo(0, NominalRecord, nchannel);
        clearStats(true);

        progressWatcher = null;

        currentRecord = null;
        currentRecordInfo = null;

        LumpSource lumpSource = new LumpSource(rawSource);

        BaselineSource baseSource = new BaselineSource(lumpSource);

        DirSource dirSource = new DirSource(baseSource);

        source = new CrowdSource(dirSource);
    }

    private void copyStats(final TapeDecoder t) {
        t.evalErrors = evalErrors;
        t.evalFixups = evalFixups;
        t.gapJunk = gapJunk;
        if (spacingHist != null) {
            if (t.spacingHist == null || t.spacingHist.length < spacingHist.length) {
                t.spacingHist = new int[spacingHist.length];
            }
            int i;
            for (i = 0; i < spacingHist.length; i++) {
                t.spacingHist[i] = spacingHist[i];
            }
            for (; i < t.spacingHist.length; i++) {
                t.spacingHist[i] = 0;
            }
        }
    }

    @Override
    public MultiSource clone() {
        TapeDecoder td = new TapeDecoder(source, datacollection, datafile);
        return td;
    }

    public Source getSource() {
        return source;
    }

    public DataFile getDataFile() {
        return datafile;
    }

    public long getStart() {
        return source.getStart();
    }

    public long getEnd() {
        return source.getEnd();
    }

    public void setStart(final long pos) {
        source.setStart(pos);
    }

    public void setEnd(final long pos) {
        source.setEnd(pos);
    }

    public long getLength() {
        return source.getLength();
    }

    public TapeParameters getTapeParameters() {
        return datafile.getTapeParameters().getTapeParameters(getPosition());
    }

    public void calibrate(final ProgressWatcher pw) throws Exception {
        calibrate(pw, null);
    }

    public void calibrate(final ProgressWatcher pw, final MultiSource transfer) throws Exception {
        if (source == null) {
            throw new Exception("Source instanceof closed");
        }
        if (progressWatcher != null) {
            throw new Exception("Operation already in progress");
        }
        progressWatcher = pw;

        int total = (int)source.getEnd();

        try {
            if (progressWatcher != null) {
                progressWatcher.reportProgress("skew", 0, total);
            }
            source.setPosition(source.getStart());
            source.resetLimit();
            SkewEval skewEval = new SkewEval(source, progressWatcher, ClumpSize, source.getTapeParameters().getWidth());
            if (progressWatcher != null) {
                progressWatcher.setPlot(skewEval);
            }
        } catch (StopException _) {
        }

        if (progressWatcher != null) {
            progressWatcher.finished();
            progressWatcher = null;
        }
    }

    public void mark(final ProgressWatcher pw) throws Exception {
        xmark(pw, null);
    }

    public void mark(final ProgressWatcher pw, final MultiSource transfer) throws Exception {
        xmark(pw, transfer);
        if (pw != null) {
            pw.finished();
        }
    }

    private void xmark(final ProgressWatcher pw, final MultiSource transfer) throws Exception {
        if (source == null) {
            throw new Error("Source instanceof closed");
        }
        if (progressWatcher != null) {
            throw new Exception("Operation already in progress");
        }

        progressWatcher = pw;
        clearStats(true);
        datafile.getRecords().clear();

        for (Source s = source; s != null; s = s.getSource()) {
            if (s instanceof BaselineSource) {
                ((BaselineSource)s).setDataPoints(null, 0);
                break;
            }
        }

        int total = (int)source.getEnd();

        try {
            if (progressWatcher != null) {
                progressWatcher.reportProgress("mark", 0, total);
            }
            source.setPosition(source.getStart());
            long gaplength = 0;
            long recStart = -1;
            long recEnd = -1;
            long lastEnd = 0;
            int clump = 0;
            int[] oldSamples = new int[nchannel];
            int hiatus = 0;
            int approxLen = 0;
            for (; !source.atLimit();) {
                int ch;
                int[] newSamples = source.getSamples();
                for (ch = 0; ch < nchannel; ch++) {
                    oldSamples[ch] = newSamples[ch];
                }

                source.advance();

                if (hiatus > 0) {
                    hiatus--;
                    ch = nchannel;
                } else {
                    newSamples = source.getSamples();
                    for (ch = 0; ch < nchannel; ch++) {
                        if (newSamples[ch] != 0 && newSamples[ch] != oldSamples[ch]) {
                            break;
                        }
                    }
                }

                if (ch < nchannel) {
                    recEnd = source.getPosition();
                    if (recStart == -1) {
                        recStart = recEnd;
                        if (recStart > Pad) {
                            recStart -= Pad;
                        } else {
                            recStart = 0;
                        }
                        approxLen = 0;
                    }
                    approxLen++;
                    hiatus = source.getTapeParameters().getWidth();
                    gaplength = 0;
                } else {
                    if (recStart != -1) {
                        TapeParameters tp = source.getTapeParameters();
                        gaplength++;
                        if (gaplength == tp.getMaxSpace()) {
                            if (recStart - lastEnd > tp.getMinGap() || approxLen > tp.getMaxJunk()) {
                                recEnd += Pad;
                                if (recEnd >= source.getEnd()) {
                                    recEnd = source.getEnd() - 1;
                                }
                                createRecord(source, recStart, recEnd, approxLen);
                                lastEnd = recEnd;
                                recStart = -1;
                                gaplength = 0;
                            } else {
                                gapJunk++;
                            }
                            recStart = -1;
                        }
                    }
                }

                if (progressWatcher != null) {
                    ++clump;
                    if (clump == ClumpSize) {
                        String s = datafile.getRecords().getCount() + " records marked";
                        progressWatcher.reportProgress(s, (int)getPosition(), total);
                        clump = 0;
                    }
                }
            }
            if (recStart != -1) {
                recEnd += Pad;
                if (recEnd >= source.getEnd()) {
                    recEnd = source.getEnd() - 1;
                }
                createRecord(source, recStart, recEnd, approxLen);
            }

            if (progressWatcher != null && clump > 0) {
                String s = datafile.getRecords().getCount() + " records marked";
                progressWatcher.reportProgress(s, (int)getPosition(), total);
                clump = 0;
            }
        } catch (StopException _) {
        }

        if (transfer instanceof TapeDecoder) {
            TapeDecoder t = (TapeDecoder)transfer;
            copyStats(t);
        }

        progressWatcher = null;
    }

    public void evaluate(final ProgressWatcher pw) throws Exception {
        evaluate(pw, null);
    }

    public void evaluate(final ProgressWatcher pw, final MultiSource transfer) throws Exception {
        if (source == null) {
            throw new Error("Source instanceof closed");
        }
        if (progressWatcher != null) {
            throw new Exception("Operation already in progress");
        }

        if (datafile.getRecords().getCount() == 0) {
            xmark(pw, transfer);
        }
        write(pw, null, false);

        if (transfer instanceof TapeDecoder) {
            TapeDecoder t = (TapeDecoder)transfer;
            copyStats(t);
        }
    }

    private void createRecord(final Source source, final long start, final long end, final int approxLen) {

        Source rawSource = source.getSource();
        for (; !(rawSource instanceof RawSource); rawSource = rawSource.getSource()) {
            ;
        }
        if (approxLen > 1 && approxLen < MaxRecord) {
            Record rec = new Record(datafile.getRecords().getCount() + 1, start, end, (((RawSource)rawSource).getFlags() & RawSource.FlagReverse) == 0);
            rec.setCount(approxLen);
            datafile.getRecords().add(rec);
        }
    }

    public RecordInfo evaluateRecord(final Record rec) throws Exception {
        return evaluateRecord(rec, false);
    }

    public RecordInfo evaluateRecord(final Record rec, final boolean setComp) throws Exception {
        boolean knewit = true;
        boolean needit = false;
        TapeParameters rtp = datafile.getTapeParameters().getTapeParameters(rec.getStart());
        for (Source s = source; s != null; s = s.getSource()) {
            if (s instanceof BaselineSource) {
                ((BaselineSource)s).setDataPoints(null, 0);
                break;
            }
        }
        RecordInfo ri = evaluateRecord(rec, source);
        if (setComp && (ri.getErrors() > 0 || ri.getMissing() > 0)) {
            knewit = rtp.getLowClip();
            needit = true;
        }
        if (rtp.getLowClip() || needit) {
            if (!knewit) {
                if (rtp.getStart() != rec.getStart() || rtp.getEnd() != rec.getEnd()) {
                    rtp = datafile.getTapeParameters().add(rec.getStart(), rec.getEnd());
                }
                for (int ch = 0; ch < nchannel; ch++) {
                    int oldclip = rtp.getClip(false, ch);
                    rtp.setClip(3 * oldclip / 8, true, ch);
                    rtp.getBaseLimit()[ch] = oldclip / 2;
                }
                int[] skewOff = ri.skewOffset();
                int min = 100;
                for (int ch = 0; ch < nchannel; ch++) {
                    int t = rtp.getSkew()[ch] + skewOff[ch];
                    if (t < min) {
                        min = t;
                    }
                }
                for (int ch = 0; ch < nchannel; ch++) {
                    rtp.getSkew()[ch] += skewOff[ch] - min;
                }
                ri = evaluateRecord(rec, source);
            }

            int[] dp = new int[ri.getLength()];
            for (int si = 0; si < dp.length; si++) {
                dp[si] = ri.getDataPoint(si);
            }
            for (Source s = source; s != null; s = s.getSource()) {
                if (s instanceof BaselineSource) {
                    ((BaselineSource)s).setDataPoints(dp, rec.getStart());
                    break;
                }
            }

            ri = evaluateRecord(rec, source);
        }
        rec.setCount(ri.getDataCount());
        rec.setErrors(ri.getErrors());
        rec.setFixups(ri.getFixups());
        if (ri.getErrors() > evalErrors) {
            evalErrors = ri.getErrors();
        }
        return ri;
    }

    private RecordInfo evaluateRecord(final Record rec, final Source source) {
        source.setPosition(rec.getStart());
        source.setLimit(rec.getEnd());

        ri.reset(rec.getStart(), (int)(rec.getEnd() - rec.getStart()));
        int[] oldSamples = new int[nchannel];
        int[] newSamples = source.getSamples();
        for (int i = 0; i < nchannel; i++) {
            oldSamples[i] = 0;
        }
        // int alternation = 0;
        for (; !source.atLimit();) {
            for (int i = 0; i < nchannel; i++) {
                if (newSamples[i] != 0) {
                    oldSamples[i] = newSamples[i];
                }
            }

            source.advance();

            ri.increment();

            newSamples = source.getSamples();
            for (int i = 0; i < nchannel; i++) {
                if (newSamples[i] != 0 && newSamples[i] != oldSamples[i]) {
                    ri.setTransition(i);
                    if (newSamples[i] > 0 && oldSamples[i] > 0 || newSamples[i] < 0 && oldSamples[i] < 0) {
                        ri.missing();
                    }
                }
            }
        }

        boolean lrcOK = true;
        for (int i = 0; i < nchannel; i++) {
            if (newSamples[i] > 0) {
                lrcOK = false;
                break;
            }
        }

        // Console.out.writeLine("-1 " + rec.getNumber());
        ri.evaluate(datafile.getTapeParameters().getTapeParameters(rec.getStart()), rec.getMotionForward(), lrcOK);
        return ri;
    }

    public void write(final ProgressWatcher pw) throws Exception {
        write(pw, null, true);
    }

    public void write(final ProgressWatcher pw, final MultiSource transfer) throws Exception {
        write(pw, null, true);
        if (transfer instanceof TapeDecoder) {
            copyStats((TapeDecoder)transfer);
        }
    }

    public void write(final ProgressWatcher pw, final PrintStream log) throws Exception {
        write(pw, log, true);
    }

    private void write(final ProgressWatcher pw, final PrintStream log, final boolean actuallyWrite) throws Exception {
        if (source == null) {
            throw new Exception("Source instanceof closed");
        }
        if (progressWatcher != null) {
            throw new Exception("Operation already in progress");
        }

        boolean setComp = !actuallyWrite;
        if (datafile.getRecords().getCount() == 0) {
            xmark(pw, null);
            setComp = true;
        }

        progressWatcher = pw;

        OutputStream outfile = null;

        if (spacingHist == null) {
            spacingHist = new int[datafile.getTapeParameters().get(0).getMaxSpace()];
        }
        clearStats(false);

        String doing = "";

        try {
            if (actuallyWrite) {
                String outfilename = datafile.getFilename() + ".bcd";
                int dot = datafile.getFilename().lastIndexOf('.');
                if (dot > 0) {
                    outfilename = datafile.getFilename().substring(0, dot) + ".bcd";
                }
                outfile = new FileOutputStream(outfilename);
                if (outfile == null) {
                    throw new Exception("Can't open output file " + outfilename);
                }
                doing = "Writing " + outfilename;
            } else {
                doing = "Evaluating";
            }

            if (progressWatcher != null) {
                progressWatcher.reportProgress(doing, 0, datafile.getRecords().getCount());
            }
            int clump = 0;

            int end = datafile.getRecords().getCount() - 1;
            int start = 0;
            int incr = 1;
            for (; start < datafile.getRecords().getCount() && !datafile.getRecords().get(start).getMotionForward(); start++) {
                end = 0;
                incr = -1;
            }
            if (start > 0) {
                start--;
            }
            for (int rn = start;; rn += incr) {
                Record rec = datafile.getRecords().get(rn);

                if (rec.getMotionForward() && incr == -1) {
                    break;
                }

                RecordInfo ri = evaluateRecord(rec, setComp);
                if (ri.getErrors() > 0) {
                    evalErrors += ri.getErrors();
                    if (log != null) {
                        log.println("Record " + rec.getNumber() + " errors " + ri.getErrors());
                    }
                }
                evalFixups += ri.getFixups();
                if (ri.getSpacing() < spacingHist.length) {
                    spacingHist[ri.getSpacing()]++;
                }

                if (outfile != null) {
                    byte[] buf = ri.getData(incr == -1);

                    if (buf != null) {
                        if (nchannel < 8) {
                            buf[0] |= 0x80;
                        }
                        outfile.write(buf, 0, buf.length);
                    }
                }

                if (rn == end) {
                    break;
                }

                if (progressWatcher != null) {
                    clump += (int)(rec.getEnd() - rec.getStart());
                    if (clump >= ClumpSize) {
                        int rnf = rn - start;
                        int total = end - start + 1;
                        if (incr < 0) {
                            rnf = -rnf;
                            total = -total;
                        }
                        String s = doing + ", " + rnf + " records";
                        progressWatcher.reportProgress(s, rnf, total);
                        clump = 0;
                    }
                }
            }
        } catch (StopException _) {
        }

        if (outfile != null) {
            outfile.close();
        }

        if (actuallyWrite) {
            datacollection.write();
        }

        if (progressWatcher != null) {
            progressWatcher.finished();
            progressWatcher = null;
        }
    }

    public void writeInfo() throws IOException {
        datacollection.write();
    }

    public void close() {
        if (source != null) {
            source.close();
        }
        source = null;
    }

    public void setPosition(final long position) {
        setPosition(position, 0);
    }

    public void setPosition(final long position, final long length) {
        Record rec = getRecord(position);
        if (rec != null) {
            if (currentRecord != rec) {
                currentRecord = rec;
                currentRecordInfo = null;
            }
            if (currentRecordInfo == null) {
                try {
                    currentRecordInfo = evaluateRecord(rec);
                } catch (Exception e) {
                    System.err.println(e);
                }
            }
        } else {
            currentRecord = null;
            currentRecordInfo = null;
        }

        source.setPosition(position);
        if (length > 0) {
            source.setLimit(position + length);
        } else {
            source.resetLimit();
        }
    }

    public void reEvaluate() {
        currentRecordInfo = null;
    }

    public long getPosition() {
        return source.getPosition();
    }

    public void advance() {
        source.advance();
    }

    public int getErrorCount() {
        setCurrent();
        if (currentRecordInfo != null) {
            return currentRecordInfo.getErrors();
        }
        return 0;
    }

    public int getErrorTotal() {
        return evalErrors;
    }

    public int getFixupTotal() {
        return evalFixups;
    }

    public int getGapJunk() {
        return gapJunk;
    }

    public int[] getSpacingHist() {
        return spacingHist;
    }

    public long getSpacing() {
        setCurrent();
        if (currentRecordInfo != null) {
            return currentRecordInfo.getSpacing();
        }
        return 0;
    }

    public long getNextError() {
        setCurrent();
        long pos = getPosition();
        if (currentRecordInfo != null) {
            int i = (int)(pos - currentRecord.getStart());
            for (; i < currentRecordInfo.getLength(); i++) {
                int st = currentRecordInfo.getStatus(i);
                if (st == 0 || (st & 1) != 0) {
                    break;
                }
            }
            for (; i < currentRecordInfo.getLength(); i++) {
                int st = currentRecordInfo.getStatus(i);
                if (st != 0 && (st & 1) == 0) {
                    break;
                }
            }
            for (; i < currentRecordInfo.getLength(); i++) {
                int st = currentRecordInfo.getDataPoint(i);
                if (st != 0) {
                    return currentRecord.getStart() + i;
                }
            }
        }
        Record rec = currentRecord;
        for (;;) {
            if (rec == null) {
                rec = getNextRecord();
            } else {
                rec = getNextRecord(rec);
            }
            if (rec == null) {
                break;
            }
            if (rec.getErrors() == 0) {
                continue;
            }
            try {
                RecordInfo ri = evaluateRecord(rec);
                for (int i = 0; i < ri.getLength(); i++) {
                    int st = ri.getStatus(i);
                    if (st != 0 && (st & 1) == 0) {
                        return rec.getStart() + i;
                    }
                }
            } catch (Exception e) {
                System.err.println(e);
            }
        }
        return getPosition();
    }

    public boolean hasActivity() {
        return true;
    }

    private void setCurrent() {
        if (currentRecord == null) {
            currentRecord = getRecord();
            currentRecordInfo = null;
            return;
        }
        long pos = getPosition();
        if (currentRecord.getStart() > pos || currentRecord.getEnd() < pos) {
            Record r = getRecord(pos);
            if (r == currentRecord) {
                return;
            }
            currentRecord = r;
            currentRecordInfo = null;
        }
    }

    public int getActivityCountHere() {
        setCurrent();
        if (currentRecordInfo == null) {
            return 0;
        }
        return currentRecordInfo.getCount((int)(getPosition() - currentRecord.getStart()));
    }

    public int getActivityLumpedHere() {
        setCurrent();
        if (currentRecordInfo == null) {
            return 0;
        }
        return currentRecordInfo.getLumped((int)(getPosition() - currentRecord.getStart()));
    }

    public int getActivityStatusHere() {
        setCurrent();
        if (currentRecordInfo == null) {
            return 0;
        }
        return currentRecordInfo.getStatus((int)(getPosition() - currentRecord.getStart()));
    }

    public boolean hasData() {
        return true;
    }

    public boolean isDataValidHere() {
        setCurrent();
        if (currentRecordInfo == null) {
            return false;
        }
        return currentRecordInfo.isDataValid((int)(getPosition() - currentRecord.getStart()));
    }

    public int getDataHere() {
        setCurrent();
        if (currentRecordInfo == null) {
            return 0;
        }
        return currentRecordInfo.getData((int)(getPosition() - currentRecord.getStart()));
    }

    public boolean recordOriented() {
        return true;
    }

    public int getNumRecords() {
        if (datafile.getRecords() != null) {
            return datafile.getRecords().getCount();
        }
        return 0;
    }

    public Record getRecord(final int index) {
        if (datafile.getRecords().getCount() > 0 && index > 0 && index <= datafile.getRecords().getCount()) {
            return datafile.getRecords().get(index - 1);
        }
        return null;
    }

    public Record getRecord() {
        return getRecord(getPosition());
    }

    public Record getRecord(final long position) {
        Record rec = currentRecord;
        int maxspace = source.getTapeParameters().getMaxSpace();
        for (;;) {
            if (rec != null && rec.getStart() - maxspace <= position && rec.getEnd() + maxspace >= position) {
                return rec;
            }
            if (datafile.getRecords().getCount() > 0) {
                if (rec != null) {
                    if (position > rec.getEnd()) {
                        Record r = getRecord(rec.getNumber() + 1);
                        if (r == null || position < r.getStart() - maxspace) {
                            return rec;
                        }
                        rec = r;
                        continue;
                    } else {
                        Record r = getRecord(rec.getNumber() - 1);
                        if (r == null || position > r.getEnd() + maxspace) {
                            return rec;
                        }
                        rec = r;
                        continue;
                    }
                }
                rec = datafile.getRecords().get(0);
                continue;
            }
            return null;
        }
    }

    public Record getNextRecord() {
        return getNextRecord(getPosition());
    }

    public Record getNextRecord(final Record record) {
        return getNextRecord(record.getEnd());
    }

    public Record getNextRecord(final long position) {
        if (datafile.getRecords().getCount() > 0) {
            for (int i = 0; i < datafile.getRecords().getCount(); i++) {
                Record rec = datafile.getRecords().get(i);
                if (rec.getStart() > position) {
                    return rec;
                }
            }
        }

        return null;
    }

    public Record getPreviousRecord() {
        return getPreviousRecord(getPosition());
    }

    public Record getPreviousRecord(final Record record) {
        return getPreviousRecord(record.getStart());
    }

    public Record getPreviousRecord(final long position) {
        if (datafile.getRecords().getCount() > 0) {
            for (int i = datafile.getRecords().getCount() - 1; i >= 0; i--) {
                Record rec = datafile.getRecords().get(i);
                if (rec.getEnd() < position) {
                    return rec;
                }
            }
        }

        return null;
    }
}
