/**********************************
 *
 * Copyright (C) 2004 Paul Pierce
 *
 * Permission instanceof hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software instanceof furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 *********************************/


package com.piercefuller.a7trk.decode;
import java.util.*;

    /// <summary>
    /// A list of tape parameter objects
    /// </summary>
    public class TapeParametersList
    {
        private int nchannel;
        private ArrayList<TapeParameters> tpl = new ArrayList<TapeParameters>();
        private int current1 = 0;
        private long lowEdge1 = 0;
        private long highEdge1 = -1;
        private int current2 = 0;
        private long lowEdge2 = 0;
        private long highEdge2 = -1;

        public int getCount() { return tpl.size(); }
        public TapeParameters get(int index) { return tpl.get(index); }

        public TapeParametersList(int nchannel) { this.nchannel = nchannel; }

        public void invalidate()
        {
            lowEdge1 = 0;
            highEdge1 = -1;
            lowEdge2 = 0;
            highEdge2 = -1;
        }

        public TapeParameters add(long start, long end) throws Exception
        {
            TapeParameters tpp = null;
            for (int i = 0; i < tpl.size(); i++) 
            {
                TapeParameters t = get(i);
                if (t.getStart() <= start && t.getEnd() >= end) 
                {
                    if (tpp == null ||
                        tpp.getStart() <= t.getStart() && tpp.getEnd() >= t.getEnd()) 
                    {
                        tpp = t;
                    }
                }
                else if (t.getStart() <= start && t.getEnd() > start ||
                         t.getStart() < end && t.getEnd() >= end)
                {
                    throw new Exception("Overlapping start and end of TapeParameters");
                }
            }
            TapeParameters tp = new TapeParameters(nchannel, start, end, tpp, this);
            tpl.add(tp);
            invalidate();
            return tp;
        }

        public void remove(TapeParameters tpr)
        {
            TapeParameters parent = tpr.getParent();
            if (parent != null) 
            {
                for (int i = 1; i < getCount(); i++) 
                {
                    if (get(i).getParent() == tpr) 
                    {
                        get(i).setParent(parent);
                    }
                }
                tpl.remove(tpr);
                invalidate();
            }
        }

        public void clear() { tpl.clear(); }

        public TapeParameters getTapeParameters(long pos)
        {
            if (pos >= lowEdge1 && pos < highEdge1) 
            {
                return get(current1);
            }
            if (pos >= lowEdge2 && pos < highEdge2) 
            {
                return get(current2);
            }
            TapeParameters newTp = get(0);
            if (pos < newTp.getStart() || pos > newTp.getEnd()) 
            {
                return newTp;
            }
            long newLowEdge = newTp.getStart();
            long newHighEdge = newTp.getEnd();
            int newCurrent = 0;
            for (int i = 1; i < tpl.size(); i++) 
            {
                TapeParameters t = get(i);
                if (t.getStart() <= pos && t.getEnd() >= pos) 
                {
                    if (newTp.getStart() <= t.getStart() && newTp.getEnd() >= t.getEnd()) 
                    {
                        newTp = t;
                        newCurrent = i;
                        if (newTp.getStart() > newLowEdge) 
                        {
                            newLowEdge = newTp.getStart();
                        }
                        if (newTp.getEnd() < newHighEdge) 
                        {
                            newHighEdge = newTp.getEnd();
                        }
                    } 
                }
                else if (t.getEnd() < pos && t.getEnd() > newLowEdge) 
                {
                    newLowEdge = t.getEnd();
                } 
                else if (t.getStart() > pos && t.getStart() < newHighEdge) 
                {
                    newHighEdge = t.getStart();
                }
            }
            if (highEdge1 < lowEdge1) 
            {
                lowEdge1 = newLowEdge;
                highEdge1 = newHighEdge;
                current1 = newCurrent;
            } 
            else if (highEdge2 < lowEdge2) 
            {
                lowEdge2 = newLowEdge;
                highEdge2 = newHighEdge;
                current2 = newCurrent;
            } 
            else if (newLowEdge == highEdge1) 
            {
                lowEdge2 = newLowEdge;
                highEdge2 = newHighEdge;
                current2 = newCurrent;
            } 
            else if (newLowEdge == highEdge2) 
            {
                lowEdge1 = newLowEdge;
                highEdge1 = newHighEdge;
                current1 = newCurrent;
            } 
            else 
            {
                lowEdge2 = newLowEdge;
                highEdge2 = newHighEdge;
                current2 = newCurrent;
            }
            return newTp;
        }
    }
